@@ -19,6 +19,7 @@ import asyncio as _asyncio
 import logging as _logging
 import inspect as _inspect
 import tempfile as _tempfile
+import urllib.parse as _urlparse
 
 import click as _click
 import asyncssh as _ssh
@@ -61,6 +62,15 @@ class TuesError(Exception):
     pass
 
 
+class TuesErrorGroup(TuesError):
+
+    def __init__(self, message, exceptions, results):
+        self.args = [message]
+        self.message = message
+        self.exceptions = exceptions
+        self.results = results
+
+
 class TuesLookupError(TuesError):
     pass
 
@@ -251,13 +261,17 @@ class Task:
         if files is None:
             files = []
 
-        if host is None and connection is None:
-            raise RuntimeError("You need to either provide a `host` or `connection` argument")
-
         if isinstance(cmd, list):
             cmd = _shlex_join(cmd)
 
-        self.host = host
+        if host is not None and not isinstance(host, Host):
+            host = Host(host)
+        elif host is None and connection is None:
+            raise RuntimeError("You need to either provide a `host` or `connection` argument")
+
+        self.host = host.name
+        self.port = host.port
+
         self.connection = connection
         self.cmd = cmd
         self.precmds = []
@@ -266,7 +280,7 @@ class Task:
         self.files = files
         self.outfile = outfile
         self.prefix = prefix
-        self.prefix_width_hint = prefix_width_hint or len(host)
+        self.prefix_width_hint = prefix_width_hint or len(self.host)
         self.user = user
         self.pty = pty
         self.input = input
@@ -282,6 +296,7 @@ class Task:
         self._stderr = None
         self.preexec_fn = preexec_fn
         self.postexec_fn = postexec_fn
+        self._session = None
 
     @property
     def stdout(self):
@@ -312,6 +327,18 @@ class Task:
 
         return self.cmdwrapper(f"{precmds}{self.cmd}")
 
+    def cleanup(self):
+        if not self._session:
+            return
+
+        if not self.pty:
+            if self.user:
+                _log.warning("Cannot terminate remote command when running as another user without pty mode, this may lead to hanging processes on the remote server")
+            else:
+                self._session.terminate()
+        else:
+            self._session.stdin.write(b"\x03")
+
 
 class BufferedIO(_tempfile.SpooledTemporaryFile):
     """ Capture output into memory or into a tempfile if output is more than 4kb
@@ -604,114 +631,118 @@ async def _redirect_io(session, stdout, stderr, sudo):
 
 async def _run(run, pm, stdout=None, stderr=None): # pylint: disable=too-many-locals,too-many-branches
     """ The real "runner" function issued for every host """
-    try:
-        env = []
-
-        # Either use an already existing connection on the `run` object
-        # or make a new one from `.host` and `.login_user`, in the later
-        # case, do *not* write it back to the run, but get rid of it as
-        # soon as we are done with the run, so we don't "leak" connections
-        # when iterating over a large number of hosts
-        if run.connection is None:
-            async with _timeout.timeout(10):
-                conn = await _ssh.connect(run.host, known_hosts=None, username=run.login_user)
-        else:
-            conn = run.connection
-
-        # Copy files to the remote environment and put their names into `TUES_FILE<x>`,
-        # even though you know the name you just put into the `files` list anyway...
-        # I have no idea why I added this environment variables to be honest...
-        for idx, f in enumerate(run.files, start=1):
-            try:
-                await _ssh.scp(f, (conn, "./"), preserve=True)
-                env.append(f"TUES_FILE{idx}=$PWD{_shlex.quote('/' + _os.path.basename(f))}")
-            except _ssh.sftp.SFTPFailure as e:
-                raise TuesError(f"Could not upload {f} to {run.host}") from e
-
-        # If there is any input to be written to the executed commands' stdin, we
-        # can write it immediately if we don't have to handle sudo situation, otherwise
-        # we need to hold back the input until we are past the sudo prompt
-        if run.input or run.stdin:
-            async def send_input():
-                if run.input:
-                    input = run.input if not run.text else run.input.encode(run.encoding, run.errors)
-                    #_log.debug("RUN INPUT %r %r", run.text, input)
-                    session.stdin.write(input)
-                    await session.stdin.drain()
-                elif isinstance(run.stdin, (str, bytes)):
-                    with open(run.stdin, "rb") as f:
-                        while True:
-                            buf = f.read(4096)
-                            if buf:
-                                session.stdin.write(buf)
-                                await session.stdin.drain()
-                            if len(buf) < 4096:
-                                break
-                elif isinstance(run.stdin, (_io.StringIO, _io.BytesIO)):
-                    input = run.stdin.getvalue()
-                    if isinstance(input, str):
-                        input = input.encode(run.encoding, run.errors)
-                    session.stdin.write(input)
-                    await session.stdin.drain()
-                session.stdin.write_eof()
-        else:
-            async def send_input():
-                pass
-
-        if run.env:
-            env.extend(f"{k}={_shlex.quote(v)}" for k, v in run.env.items())
-
-        stdout, stderr, sudo, cleanup = await _prepare_io(
-            run,
-            stdout,
-            stderr,
-            env,
-            pm,
-            send_input,
-        )
+    env = []
 
+    # Either use an already existing connection on the `run` object
+    # or make a new one from `.host` and `.login_user`, in the later
+    # case, do *not* write it back to the run, but get rid of it as
+    # soon as we are done with the run, so we don't "leak" connections
+    # when iterating over a large number of hosts
+    if run.connection is None:
         try:
-            if run.preexec_fn:
-                run.preexec_fn(run)
+            async with _timeout.timeout(10):
+                conn = await _ssh.connect(run.host, known_hosts=None, username=run.login_user, port=run.port or ())
+        except Exception as e:
+            raise TuesError(f"Could not connect to {run.host}: {e!r}") from e
+    else:
+        conn = run.connection
 
-            cmd = run.build_cmd()
-            _log.debug("Running %r", cmd)
-            _chan, session = await conn.create_session(
-                _ssh.SSHClientProcess,
-                cmd,
-                term_type=_os.environ.get("TERM"),
-                request_pty=run.pty,
-                env=env,
-                # If we don't "force" them to None, some kind of autodetection will take place
-                # and we'll end up receiving `str` objects in some scenarios.
-                encoding=None,
-                errors=None,
-            )
-            await _redirect_io(session, stdout, stderr, sudo)
-            if not sudo:
-                await send_input()
+    # Copy files to the remote environment and put their names into `TUES_FILE<x>`,
+    # even though you know the name you just put into the `files` list anyway...
+    # I have no idea why I added this environment variables to be honest...
+    for idx, f in enumerate(run.files, start=1):
+        try:
+            await _ssh.scp(f, (conn, "./"), preserve=True)
+            env.append(f"TUES_FILE{idx}=$PWD{_shlex.quote('/' + _os.path.basename(f))}")
+        except _ssh.sftp.SFTPFailure as e:
+            raise TuesError(f"Could not upload {f} to {run.host}") from e
 
-            completed = await session.wait()
+    # If there is any input to be written to the executed commands' stdin, we
+    # can write it immediately if we don't have to handle sudo situation, otherwise
+    # we need to hold back the input until we are past the sudo prompt
+    if run.input or run.stdin:
+        async def send_input():
+            if run.input:
+                input = run.input if not run.text else run.input.encode(run.encoding, run.errors)
+                session.stdin.write(input)
+                await session.stdin.drain()
+            elif isinstance(run.stdin, (str, bytes)):
+                with open(run.stdin, "rb") as f:
+                    while True:
+                        buf = f.read(4096)
+                        if buf:
+                            session.stdin.write(buf)
+                            await session.stdin.drain()
+                        if len(buf) < 4096:
+                            break
+            elif isinstance(run.stdin, (_io.StringIO, _io.BytesIO)):
+                input = run.stdin.getvalue()
+                if isinstance(input, str):
+                    input = input.encode(run.encoding, run.errors)
+                session.stdin.write(input)
+                await session.stdin.drain()
+            session.stdin.write_eof()
+    else:
+        async def send_input():
+            pass
 
-            run.returncode = completed.returncode
+    if run.env:
+        env.extend(f"{k}={_shlex.quote(v)}" for k, v in run.env.items())
 
-            if run.postexec_fn:
-                run.postexec_fn(run)
+    stdout, stderr, sudo, cleanup = await _prepare_io(
+        run,
+        stdout,
+        stderr,
+        env,
+        pm,
+        send_input,
+    )
 
-            return run
-        finally:
-            if conn:
-                if run.files:
-                    await conn.run(b"rm -f " + b" ".join([_shlex.quote(_os.path.basename(_)).encode() for _ in run.files]))
+    try:
+        if run.preexec_fn:
+            run.preexec_fn(run)
 
-                if not run.connection:
-                    conn.close()
+        cmd = run.build_cmd()
+        _log.debug("Running %r", cmd)
+        _chan, session = await conn.create_session(
+            _ssh.SSHClientProcess,
+            cmd,
+            term_type=_os.environ.get("TERM"),
+            request_pty=run.pty,
+            env=env,
+            # If we don't "force" them to None, some kind of autodetection will take place
+            # and we'll end up receiving `str` objects in some scenarios.
+            encoding=None,
+            errors=None,
+        )
+        await _redirect_io(session, stdout, stderr, sudo)
+        if not sudo:
+            await send_input()
 
-            for cb in cleanup:
-                cb()
-    except Exception:
-        _log.error("Error during run", exc_info=True)
-        raise
+        # Since we won't receive the KeyboardInterrupt here, we need to store the
+        # session in a way that will allow us to call `task.session.terminate()`
+        # from a location that does see it.
+        run._session = session
+        completed = await session.wait()
+        run._session = None
+
+        run.returncode = completed.returncode
+
+        if run.postexec_fn:
+            run.postexec_fn(run)
+
+        return run
+    finally:
+        if conn:
+            if run.files:
+                await conn.run(b"rm -f " + b" ".join([_shlex.quote(_os.path.basename(_)).encode() for _ in run.files]))
+
+            # We used a connection from the run object, don't close it
+            if not run.connection:
+                conn.close()
+
+        for cb in cleanup:
+            cb()
 
 
 async def _wait_with_concurrency(tasks, pool_size):
@@ -730,13 +761,19 @@ async def run_tasks(tasks, pm=_PM, stdout=None, stderr=None, pool_size=1): # pyl
 
     (_done, _pending) = await _wait_with_concurrency(tasks, pool_size)
 
+    excs = []
+    results = []
     for d in _done:
         exc = d.exception()
         if exc:
-            if isinstance(exc, OSError):
-                print(exc.args)
-            else:
-                _tb.print_exception(type(exc), exc, exc.__traceback__)
+            excs.append(exc)
+        else:
+            results.append(d.result())
+
+    if excs:
+        raise TuesErrorGroup("Errors encountered while running tasks with concurrency", excs, results)
+
+    return results
 
 
 def _prepare_output_dir(path, strategy):
@@ -769,6 +806,17 @@ def _prepare_output_dir(path, strategy):
         _os.mkdir(path)
 
 
+class Host:
+
+    def __init__(self, spec):
+        if isinstance(spec, tuple):
+            self.name, self.port = spec
+        elif isinstance(spec, str):
+            parsed = _urlparse.urlparse(f"ssh://{spec}")
+            self.name = parsed.hostname
+            self.port = parsed.port
+
+
 def run(
     hosts,
     cmd,
@@ -912,13 +960,15 @@ def run(
                 is created and used internally, you don't need to `await` anything.
 
     """
-    if isinstance(hosts, str):
+    if isinstance(hosts, (str, tuple)):
         hosts = [hosts]
 
+    hosts = [Host(_) for _ in hosts]
+
     if output_dir:
         _prepare_output_dir(output_dir, output_dir_strategy)
 
-    host_width = max(len(_) for _ in hosts)
+    host_width = max(len(_.name) for _ in hosts)
 
     tasks = [
         Task(
@@ -926,7 +976,7 @@ def run(
             cmd=cmd,
             login_user=login_user,
             files=files,
-            outfile=_os.path.join(output_dir, host + ".log") if output_dir else None,
+            outfile=_os.path.join(output_dir, host.name + ".log") if output_dir else None,
             prefix=prefix,
             prefix_width_hint=host_width if align_prefix else None,
             user=user,
@@ -950,17 +1000,21 @@ def run(
         except RuntimeError:
             loop = _asyncio.get_event_loop()
 
-    if pool_size > 1:
-        if check:
-            raise TuesError("Aborting on errors is not supported when running with a pool size > 1")
-        result = loop.run_until_complete(
-            run_tasks(tasks, pm, stdout, stderr, pool_size),
-        )
-    else:
+    try:
+        if pool_size > 1:
+            if check:
+                raise TuesError("Aborting on errors is not supported when running with a pool size > 1")
+            result = loop.run_until_complete(
+                run_tasks(tasks, pm, stdout, stderr, pool_size),
+            )
+        else:
+            for task in tasks:
+                result = loop.run_until_complete(_run(task, pm, stdout, stderr))
+                if check and result.returncode > 0:
+                    raise TuesTaskError(task)
+    except KeyboardInterrupt as e:
         for task in tasks:
-            result = loop.run_until_complete(_run(task, pm, stdout, stderr))
-            if check and result.returncode > 0:
-                raise TuesTaskError(task)
+            task.cleanup()
 
     if len(tasks) > 1:
         return tasks
